///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2002 Jason Baldridge and Gann Bierner
// Modified for supertagging where indicated
// Changes (C) 2004 Richard Moot (Richard.Moot@labri.fr)
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////////////////////////////////////////////////////////////////////////

import opennlp.common.morph.*;
import opennlp.common.util.*;
import opennlp.maxent.*;

import opennlp.tools.util.BeamSearchContextGenerator;

import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * A context generator for the supertagger.
 * Modified from the POS context generator by Gann Bierner     20040203 RM
 *
 * @author      Gann Bierner
 * @version     $Revision: 1.2 $, $Date: 2006/02/27 14:31:37 $
 */

public class STContextGenerator extends TriContextGenerator implements BeamSearchContextGenerator{
    
    /**
     * Either: (rare?, [words, tags, supertags, position])
     * or    : [words, tags, supertags, position]
     *
     * right now this will be very inefficient for a linked list.
     *
     * Context information is now the following:
     *
     *                -2   -1    0   +1   +2
     * word           pp    p    w    n   nn
     *              [   cp    ]    [   cn    ]
     * tag            tt    t    o    u   uu
     *              [   ctp   ]    [   ctn   ]
     * asciitag       xx    x    a    y   yy
     *              [   atp   ]    [   atn   ]
     * supertag       ss    s    - (  r   rr )
     *              [   stp   ]
     * goal           ff    f    - (  e   ee )
     * leftarg        ll    l    - (  m   mm )
     * rightarg       gg    g    - (  h   hh )
     * lp            llp   lp    - ( lq  llq )
     * rp            rrp   rp    - ( rq  rrq )
     *
     * Commented out the `crossed' combined features, since they
     * explode the number of predicates without helping much. Also
     * removed the `stoplist' to try training with a manually cleaned
     * dataset.                                                 20041209 RM
     * Add more combined features, this time combinations of
     * different features, eg. word-tag or supertag-tag         20040408 RM
     * Added features which are the POS tags stripped of the trailing
     * number ie. `ww' instead of `ww9'                         20040327 RM
     * Also added combined supertag information                 20040327 RM
     *
     * Added combined features                                  20040319 RM
     *
     * (feature names in parenthesis require reimplementation of beam
     *  search using dynamic programming)                       20040203 RM
     * 
     * Corrected mistake here                                   20040209 RM
     */

    public ArrayList auxGetContext(int index, Object[] words, Object[] tags, String[] stags, Object[] additionalContext) {

	String pw, ppw, nw, nnw, pv, ppv, np, nnp, tmp;
	int i;

	pw=ppw=nw=nnw=pv=ppv=np=nnp=null;
//        	pf=ppf=pl=ppl=pr=ppr=plp=pplp=prp=pprp=0;

	/* features for the previous/next verb and preposition, which
	   should help correct supertag assignment      20050101 RM
	   TODO: what about previous wh-word, eg. VNW13-15? */

	ArrayList e = super.auxGetContext(index, words, tags, stags, additionalContext);

	// previous preposition and verb

	for (i = 0; i < index; i++) {
	    tmp = getSimpleTag((String) tags[i]);
	    if (tmp.equals("ww")) {
		ppw = pw;
		pw = (String) words[i];
	    } else if (tmp.equals("vz")) {
		ppv = pv;
		pv = (String) words[i];
	    }
	}

	// next preposition and verb

	for (i = (words.length-1); i > index; i--) {
	    tmp = getSimpleTag((String) tags[i]);
	    if (tmp.equals("ww")) {
		nnw = nw;
		nw = (String) words[i];
	    } else if (tmp.equals("vz")) {
		nnp = np;
		np = (String) words[i];
	    }
	}

	// add preposition/verb information

	if(pw!=null) {
	    e.add("pw="+pw);
	}
	if(ppw!=null) {
	    e.add("ppw="+ppw);
	}
	if(nw!=null) {
	    e.add("nw="+nw);
	}
	if(nnw!=null) {
	    e.add("nnw="+nnw);
	}
	if(pv!=null) {
	    e.add("pv="+pv);
	}
	if(ppv!=null) {
	    e.add("ppv="+ppv);
	}
	if(np!=null) {
	    e.add("np="+np);
	}
	if(nnp!=null) {
	    e.add("nnp="+nnp);
	}
	
        return e;  
    }

    public static void main(String[] args) {
	  
	  STContextGenerator gen = new STContextGenerator();

	  // zoals/vg2/1860 al/bw/1484 gezegd/ww7/229 we/vnw1/148 zijn/ww2/366 dan/bw/1484 toe/vz2/1001 aan/vz1/1001 agendapunt/n1/148 negen/tw1/495 de/lid/255 mededelingen/n3/148
    
	  String[] lex =  {"zoals", "al", "gezegd", "we", "zijn", "dan", "toe", "aan", "agendapunt", "negen"};
	  String[] tags = {"vg2",   "bw", "ww7",   "vnw1", "ww2", "bw", "vz2", "vz1", "n1", "tw1"};
	  String[] stags = {"1860", "1484", "229", "148", "366", "1484", "1001", "1001", "148", "495"};

	  Pair[] p = mergeSequences(lex, tags);
	  String[] ans1 = gen.getContext(2, p, stags, null);
	  String[] ans2 = gen.getContext(4, p, stags, null);

	  for(int i=0; i<ans1.length; i++)
	      System.out.println(ans1[i]);
	  System.out.println();
	  for(int i=0; i<ans2.length; i++)
	      System.out.println(ans2[i]);
    }

}



