///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2002 Jason Baldridge and Gann Bierner
// Modified for supertagging
// Changes (C) 2004 Richard Moot (Richard.Moot@labri.fr)
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////////////////////////////////////////////////////////////////////////

import opennlp.maxent.*;

import java.io.*;
import java.util.*;

/**
 * An event generator for the maxent Supertagger.
 *
 * @author      Gann Bierner
 * @version     $Revision: 1.1.1.1 $, $Date: 2005/02/02 18:08:44 $
 */

public class STEventCollector implements EventCollector {

    private BufferedReader br;
    private AbstractContextGenerator cg;
    
    public STEventCollector(Reader data, AbstractContextGenerator gen) {
	br = new BufferedReader(data);
	cg = gen;
    }

    private static Triple split(String s) {
        // data are now assumed to be of the form WORD/TAG/SUPERTAG
        // if not, an error message is produced, also the return type
        // is changed to be of the form pair(pair(WORD,TAG),SUPERTAG) 
        //                                                20040202 RM
	int split1 = s.indexOf("/");
	int split2 = s.lastIndexOf("/");
	if (split1 == -1 || split2 == -1 || split1 == split2) {
	    System.out.println("There is a problem in your training data: "
			       + s
			       + " does not conform to the format WORD/TAG/SUPERTAG.");
	    if (split1 == split2) {
		return new Triple(s.substring(0,split1), s.substring(split1+1), "UNKNOWN");
	    } else {
		return new Triple(s, "UNKNOWN", "UNKNOWN");
	    }
	} else {
	    return new Triple (s.substring(0, split1), s.substring(split1+1,split2), s.substring(split2+1));
	}
    }
    
    // again the pair is changed so that the first element is a pair
    // of ArrayLists, the first representing the words (tokens) and
    // the second the tags                                20040202 RM

    public static Pair convertAnnotatedString(String s) {
	ArrayList tokentags = new ArrayList();
	ArrayList outcomes = new ArrayList();
	StringTokenizer st = new StringTokenizer(s);
	while(st.hasMoreTokens()) {
	    Triple t = split(st.nextToken());
	    tokentags.add(new Pair(t.a,t.b));
	    outcomes.add(t.c);
	}
	return new Pair(tokentags, outcomes);
    }

    public static Triple convertAnnotatedStringT(String s) {
	ArrayList tokens = new ArrayList();
	ArrayList tags = new ArrayList();
	ArrayList outcomes = new ArrayList();
	StringTokenizer st = new StringTokenizer(s);
	while(st.hasMoreTokens()) {
	    Triple t = split(st.nextToken());
	    tokens.add(t.a);
	    tags.add(t.b);
	    outcomes.add(t.c);
	}
	return new Triple(tokens, tags, outcomes);
    }


    private Set getFrequent(BufferedReader br) {
	HashMap map = new HashMap();

	try {
	    for(String s = br.readLine(); s!=null; s=br.readLine()) {
		StringTokenizer st = new StringTokenizer(s);
		while(st.hasMoreTokens()) {
		    String tok = (String) split(st.nextToken()).a;
		    Counter c = (Counter) map.get(tok);
		    if(c!=null)
			c.increment();
		    else
			map.put(tok, new Counter());
		}
	    }
	} catch (IOException e) {
	    e.printStackTrace(); 
	}

	HashSet set = new HashSet();
	for(Iterator i=map.entrySet().iterator(); i.hasNext();) {
	    Map.Entry entry = (Map.Entry) i.next();
	    if(((Counter) entry.getValue()).passesCutoff(5))
		set.add(entry.getKey());
	}

	return set;
    }
    
    /** 
     * Builds up the list of features using the Reader as input.  For now, this
     * should only be used to create training data.
     */

    public Event[] getEvents() {
	return getEvents(false);
    }

    public Event[] getEvents(boolean evalMode) {

	ArrayList elist = new ArrayList();
	int numMatches;

	try {
	    String s = br.readLine();
	    
	    while (s != null) {
		Triple t = convertAnnotatedStringT(s);
		ArrayList tokens = (ArrayList) t.a;
		ArrayList tags = (ArrayList) t.b;
		ArrayList outcomes = (ArrayList) t.c;
		ArrayList stags = new ArrayList();
		
		for (int i=0; i<tokens.size(); i++) {
		    String[] context = cg.getContext(i,tokens.toArray(),tags.toArray(),(String[]) stags.toArray(new String[stags.size()]),null);
		    Event e = new Event((String)outcomes.get(i), context);
		    stags.add(outcomes.get(i));
		    elist.add(e);
		}
		s = br.readLine();
	    }
	} catch (Exception e) { 
	    e.printStackTrace(); 
	}

	Event[] events = new Event[elist.size()];
        for(int i=0; i<events.length; i++)
            events[i] = (Event)elist.get(i);
 
        return events;
    }

    // replaced example in main by a CGN example           20040202 RM

    public static void main(String[] args) {
	String data = "uhm/tsw/83 zullen/ww2/12 we/vnw1/83 ze/vnw3/83 paginagewijs/adj9/12 afhandelen/ww4/765 ?/let/634";
	EventCollector ec = new STEventCollector(new StringReader(data),
						  new TriContextGenerator());
	Event[] events = ec.getEvents();
	for(int i=0; i<events.length; i++) {
	    System.out.println(events[i].getOutcome());
	}
    }
    
}
