///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2002 Jason Baldridge and Gann Bierner
// Modified for supertagging where indicated
// Changes (C) 2004 Richard Moot (Richard.Moot@labri.fr)
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////////////////////////////////////////////////////////////////////////

import opennlp.maxent.*;

import opennlp.tools.util.BeamSearchContextGenerator;


import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * A context generator for the supertagger.
 * Modified from the POS context generator by Gann Bierner     20040203 RM
 *
 * @author      Gann Bierner
 * @version     $Revision: 1.3 $, $Date: 2006/03/15 17:12:18 $
 */

public class TriContextGenerator extends AbstractContextGenerator implements BeamSearchContextGenerator {

    // pattern for obtaining the letter-only part of a POS tag

    public static Pattern pattern = Pattern.compile("^([^0-9]+)([0-9]*)$");

    // variables used in subclasses of the current class

    protected String sprev = null;
    protected String sprevprev = null;

    protected String nnext = null;
    protected String tagnnext = null;
    protected String tagnnexta = null;

    /**
     *  getSimpleTag generates a more general tag by stripping of the
     *  number distinguishing similar tags
     *  in some cases, as in the cgntoform program, we back off to
     *  a tag of a different category, for example VNW11 is a possesive
     *  determiner and therefore assigned DET 
     */

    public String getSimpleTag(String s) {

	Matcher matcher = pattern.matcher(s);
	int i = 0;

	if (matcher.find()) {

	    if (matcher.start(2) < matcher.end(2)) {
		i = Integer.parseInt(matcher.group(2));
	    }

	    s = matcher.group(1);

	    if (s.compareTo("ww") == 0) {
		if (i == 6 || i == 9 || i == 12) {
		    return "vnw";
		} else if (i == 5 || i == 8 || i == 11) {
		    return "adj";
		}
	    } else if (s.compareTo("adj") == 0) {
		if (i > 3 && i < 7) {
		    return "n";
		}
	    } else if (s.compareTo("vnw") == 0) {
		if (i == 9 || i == 11 || i == 12 || i == 18 || i == 21 || i == 24 || i == 27) {
		    return "det";
		} else if (i == 26) {
		    return "adj";
		} else if (i == 16 || i == 25) {
		    return "n";
		} else if (i > 12 && i < 18 && i != 16) {
		    return "wh";
		}
	    }
	}
	
	return s;
    }
    
    /**
     * Either: (rare?, [words, tags, supertags, position])
     * or    : [words, tags, supertags, position]
     *
     * right now this will be very inefficient for a linked list.
     *
     * Context information is now the following:
     *
     *                -2   -1    0   +1   +2
     * word           pp    p    w    n   nn
     *              [   cp    ]    [   cn    ]
     *                    [     cpn    ]                      
     * tag            tt    t    o    u   uu
     *              [   ctp   ]    [   ctn   ]
     *                    [     tpn    ]
     * asciitag       xx    x    a    y   yy
     *              [   atp   ]    [   atn   ]
     *                    [     apn    ]
     * supertag       ss    s    - (  r   rr )
     *              [   stp   ]
     * goal           ff    f    - (  e   ee )
     * leftarg        ll    l    - (  m   mm )
     * rightarg       gg    g    - (  h   hh )
     * lp            llp   lp    - ( lq  llq )
     * rp            rrp   rp    - ( rq  rrq )
     *
     * Commented out the `crossed' combined features, since they
     * explode the number of predicates without helping much. Also
     * removed the `stoplist' to try training with a manually cleaned
     * dataset.                                                 20041209 RM
     * Add more combined features, this time combinations of
     * different features, eg. word-tag or supertag-tag         20040408 RM
     * Added features which are the POS tags stripped of the trailing
     * number ie. `ww' instead of `ww9'                         20040327 RM
     * Also added combined supertag information                 20040327 RM
     *
     * Added combined features                                  20040319 RM
     *
     * (feature names in parenthesis require reimplementation of beam
     *  search using dynamic programming)                       20040203 RM
     * 
     * Corrected mistake here                                   20040209 RM
     */

    public ArrayList auxGetContext(int index, Object[] words, Object[] tags, String[] stags, Object[] additionalContext) {

	String cnext, cprev, ctnext, ctprev, atnext, atprev, stprev;
	String next, lex, prev, prevprev, cpn, tpn, apn;
	String tagprev, tagprevprev, tag, tagnext;
	String tagpreva, tagprevpreva, taga, tagnexta;

	cnext=cprev=sprev=sprevprev=stprev=null;
	ctnext=ctprev=tagprev=tagprevprev=tag=tagnext=null;
	tagpreva=tagprevpreva=taga=tagnexta=null;
	cpn=tpn=apn=next=lex=prev=prevprev=atnext=atprev=null;

	lex = (String) words[index];
	tag = (String) tags[index];

	taga = getSimpleTag(tag);

	/* features for the previous/next verb and preposition, which
	   should help correct supertag assignment      20050101 RM
	   TODO: what about previous wh-word, eg. VNW13-15? */

	if (words.length>index+1) {
	    next = (String) words[index+1];
	    tagnext = (String) tags[index+1];
	    tagnexta = getSimpleTag(tagnext);

	    if (words.length>index+2) {
		nnext = (String) words[index+2];
		tagnnext = (String) tags[index+2];
		tagnnexta = getSimpleTag(tagnnext);
		cnext = next+"_"+nnext;
		ctnext = tagnext + "_" + tagnnext;
		atnext = tagnexta + "_" + tagnnexta;
	    } else {
		nnext = "*SE*"; // Sentence End
	        tagnnext = "*SE*";
	    }
	}
	else {
	    next = "*SE*"; // Sentence End
	    tagnext = "*SE*";
	}
	    
	if(index-1>=0) {
	    prev = (String) words[index-1];
	    tagprev = (String) tags[index-1];
	    tagpreva = getSimpleTag(tagprev);
	    sprev = (String) stags[index-1];

	    // before/after combination

	    if (words.length>index+1) {
		cpn = prev+"_"+next;
		tpn = tagprev+"_"+tagnext;
		apn = tagpreva+"_"+tagnexta;
	    }

	    if(index-2>=0) {
		prevprev = (String) words[index-2];
		tagprevprev = (String) tags[index-2];
		tagprevpreva = getSimpleTag(tagprevprev);
		sprevprev = (String) stags[index-2];

		cprev = prevprev + "_" + prev;
		ctprev = tagprevprev + "_" + tagprev;
		atprev = tagprevpreva + "_" + tagpreva;
		stprev = sprevprev + "_" + sprev;
	    } else {
		prevprev = "*SB*";    // Sentence Beginning
		tagprevprev = "*SB*"; // Sentence Beginning
		sprevprev = "*SB*";   // Sentence Beginning	       
	    }
	} else {
	    prev = "*SB*";    // Sentence Beginning
	    tagprev = "*SB*"; // Sentence Beginning
	    sprev = "*SB*";   // Sentence Beginning
	}

	ArrayList e = new ArrayList();

	// add the word itself

	e.add("w="+lex);

	if (Character.isUpperCase(lex.charAt(0))) {
	    e.add("c");
	}

	if(tag!=null) {
	    e.add("o="+tag);
	}
	if(taga!=null) {
	    e.add("a="+taga);
	}
	// add the words and pos's of the surrounding context
	if(prev!=null) {
	    e.add("p="+prev);
	    if (tagprev!=null) {
		e.add("t="+tagprev);
	    }
	    if (tagpreva!=null) {
		e.add("x="+tagpreva);
	    }
	    if (sprev!=null) {
		e.add("s="+sprev);
	    }
	    
	    if(cpn!=null) {
		e.add("cpn="+cpn);
	    }
	    if(tpn!=null) {
		e.add("tpn="+tpn);
	    }
	    if(apn!=null) {
		e.add("apn="+apn);
	    }

	    if(prevprev!=null) {
		e.add("pp="+prevprev);
		if(tagprevprev!=null) {
		    e.add("tt="+tagprevprev);
		}
		if(tagprevpreva!=null) {
		    e.add("xx="+tagprevpreva);
		}
		if(sprevprev!=null) {
		    e.add("ss="+sprevprev);
		}
		if(cprev!=null) {
		    e.add("cp="+cprev);
                }
		if(ctprev!=null) {
		    e.add("ctp="+ctprev);
                }
		if(atprev!=null) {
		    e.add("atp="+atprev);
                }
		if(stprev!=null) {
		    e.add("stp="+stprev);
		}
	    }
	}

	if (next!=null) {
	    e.add("n="+next);
	    if(tagnext!=null) {
		e.add("u="+tagnext);
	    }
	    if(tagnexta!=null) {
		e.add("y="+tagnexta);
	    }

	    if (nnext!=null) {
		e.add("nn="+nnext);
		if(tagnnext!=null) {
		    e.add("uu="+tagnnext);
		}
		if(tagnnexta!=null) {
		    e.add("yy="+tagnnexta);
		}
		if(cnext!=null) {
		    e.add("cn="+cnext);
		}
		if(ctnext!=null) {
		    e.add("ctn="+ctnext);
		}
		if(atnext!=null) {
		    e.add("atn="+atnext);
		}
	    }
	}
	
        return e;  
    }

    public static void main(String[] args) {
	  
	  TriContextGenerator gen = new TriContextGenerator();

	  String line = null;
	  String[] poswords = null;
	  String[] words = null;
	  String[] pos = null;
	  String[] st = null;
	  String[] item = new String[3];

	  try {
	      if (args.length >= 2) {

		BufferedReader inf = new BufferedReader(new FileReader(args[0]));
		PrintWriter outf = new PrintWriter(new FileWriter(args[1]));

		while ((line=inf.readLine()) != null) {

		    poswords = null;
		    poswords = line.split(" ");
		    System.err.println(poswords.length);
		    words = null;
		    words = new String[poswords.length];
		    pos = null;
		    pos = new String[poswords.length];
		    st = null;
		    st = new String[poswords.length];
		    for (int i=0; i < poswords.length; i++) {
			item = poswords[i].split("/");
			words[i] = item[0];
			pos[i] = item[1];
			st[i] = item[2];
		    }
	  // zoals/vg2/1860 al/bw/1484 gezegd/ww7/229 we/vnw1/148 zijn/ww2/366 dan/bw/1484 toe/vz2/1001 aan/vz1/1001 agendapunt/n1/148 negen/tw1/495 de/lid/255 mededelingen/n3/148

		    Pair[] p = mergeSequences(words, pos);
	  
		    for (int i=0; i < poswords.length; i++) {

			String supertag = st[i];
			String[] ctx = gen.getContext(i, p, st, null);

			outf.print(supertag);
			for(int j=0; j<ctx.length; j++) {
			    outf.print(" ");
			    outf.print(ctx[j]);
			}
			outf.println();
		    }
		}
		outf.close();
		inf.close();

	      } else {
		System.out.println("Usage: java EvaluationME infile outfile model");
	      }
	  } catch (IOException e) {
	      System.err.println(e);
	      System.exit(1);
	  }
    }

}



